import abc
import itertools

class Miner(abc.ABC):

    @abc.abstractmethod
    def load(self):
        pass

    @abc.abstractmethod
    def save(self):
        pass

    @abc.abstractmethod
    def mine(self, log):
        pass

class LogSkeleton(Miner):

    def load(self):
        pass

    def save(self):
        pass

    def mine(self, log):
        # Augment the TraceLog object with unique start and end activities
        tl = log.augment()

        # Step 1: Mine equivalence relationship
        R_eq = tl.equivalence()

        # Step 2: Mine always-after relationship
        R_aa = tl.always_after()

        # Step 3: Mine always-before relationship
        R_ab = tl.always_before()

        # Step 4: Mine never-together relationship
        R_nt = tl.never_together()

        # Step 5: Mine never-together relationship
        R_df = set((tl.follows()).keys())
        
        # Step 6: Statistics
        C_df = tl.follows()
        C_sum = tl.sum_counter()
        C_max = tl.max_counter()
        C_min = tl.min_counter()

        return (R_eq, R_aa, R_ab, R_nt, R_df), (C_df, C_sum, C_max, C_min)



import os

import pytest

from skelevision import follows, successors, predecessors

from sortedcontainers import SortedSet

HERE = os.path.dirname(os.path.abspath(__file__))
DATA = os.path.join(HERE, "datasets")


def test_follows():
    t1 = ("a1","a2","a4","a5","a6","a2")
    target = {
        ("a1", "a2"): 1,
        ("a2", "a4"): 1,
        ("a4", "a5"): 1,
        ("a5", "a6"): 1,
        ("a6", "a2"): 1
    }
    f = follows(t1)

    for k, v in f.items():
        assert k in target
        assert target[k] == v

    for k, v in target.items():
        assert k in f
        assert f[k] == v

def test_follows_distance_2():
    t1 = ("a1","a2","a4","a5","a6","a2")
    target = {
        ("a1", "a4"): 1,
        ("a2", "a5"): 1,
        ("a4", "a6"): 1,
        ("a5", "a2"): 1
    }
    f = follows(t1, distance=2)

    for k, v in f.items():
        assert k in target
        assert target[k] == v

    for k, v in target.items():
        assert k in f
        assert f[k] == v


def test_follows_loop_len_1():
    t2 = ("a1","a2","a4","a5","a6","a2", "a2", "a2")
    target = {
        ("a1", "a2"): 1,
        ("a2", "a4"): 1,
        ("a4", "a5"): 1,
        ("a5", "a6"): 1,
        ("a6", "a2"): 1,
        ("a2", "a2"): 2
    }
    f = follows(t2, distance=1)

    for k, v in f.items():
        assert k in target
        assert target[k] == v

    for k, v in target.items():
        assert k in f
        assert f[k] == v

def test_follows_exception_no_integer_distance():
    t1_list = ["a1","a2","a4","a5","a6","a2"]
    with pytest.raises(ValueError):
        follows(t1_list)

def test_follows_exception_no_integer_distance():
    t1 = ("a1","a2","a4","a5","a6","a2")
    with pytest.raises(ValueError):
        follows(t1, distance=1.2)

def test_follows_exception_no_positive_integer():
    t1 = ("a1","a2","a4","a5","a6","a2")
    with pytest.raises(ValueError):
        follows(t1, distance=-3.3)

def test_successors():
    t1 = ("a1","a2","a4","a5","a6","a2")
    target = {
        "a1": SortedSet({"a2", "a4", "a5", "a6"}),
        "a2": SortedSet({"a2", "a4", "a5", "a6"}),
        "a4": SortedSet({"a2", "a5", "a6"}),              
        "a5": SortedSet({"a2", "a6"}),
        "a6": SortedSet({"a2"}),
    }
    assert successors(t1) == target

def test_predecessor():
    t1 = ("a1","a2","a4","a5","a6","a2")
    target = {
        "a2": SortedSet({"a1", "a2", "a4", "a5", "a6"}),
        "a6": SortedSet({"a1", "a2", "a4", "a5"}),
        "a5": SortedSet({"a1", "a2", "a4"}),
        "a4": SortedSet({"a1", "a2"}),
    }
    assert predecessors(t1) == target



import os

import pytest

from skelevision import TraceLog, IllegalLogAction

HERE = os.path.dirname(os.path.abspath(__file__))
DATA = os.path.join(HERE, "datasets")


class TestTraceLog(object):
    def test_constructor(self):
        d = {("a", "b", "c"): 2, ("a", "f"): 1}
        tl = TraceLog(d)

        for k, v in d.items():
            assert k in tl
            assert tl[k] == v

        for k, v in tl.items():
            assert k in d
            assert d[k] == v

    def test_labels(self):
        d = {("a", "b", "c"): 2, ("a", "f"): 1}
        labels = ["a", "b", "c", "f"]
        tl = TraceLog(d)

        for a in tl.labels:
            assert a in labels

        for a in labels:
            assert a in tl.labels

    def test_follows(self):
        d = {("a", "b", "c"): 2, ("a", "f"): 1}
        s = {("a", "b"): 2, ("b", "c"): 2, ("a", "f"): 1}
        tl = TraceLog(d)
        s_tl = tl.follows()

        for k, v in s.items():
            assert k in s_tl
            assert s_tl[k] == v

        for k, v in s_tl.items():
            assert k in s
            assert s[k] == v

    def test_follows_L1(self):
        tl = TraceLog.from_txt(os.path.join(DATA, "L1.txt"))
        tl = tl.augment()

        target = {
            ("[>", "a1"): 20,
            ("a1", "a2"): 10,
            ("a1", "a3"): 3,
            ("a1", "a4"): 7,

            ("a2", "a4"): 13,
            ("a2", "a5"): 7,

            ("a3", "a4"): 8,
            ("a3", "a5"): 6,

            ("a4", "a2"): 7,
            ("a4", "a3"): 6,
            ("a4", "a5"): 21,
            
            ("a5", "a6"): 14,
            ("a5", "a7"): 9,
            ("a5", "a8"): 11,

            ("a6", "a2"): 3,
            ("a6", "a3"): 5,
            ("a6", "a4"): 6,

            ("a7", "[]"): 9,
            ("a8", "[]"): 11,

        }

        tl_follows = tl.follows()

        for k, v in target.items():
            assert k in tl_follows
            assert tl_follows[k] == v

        for k, v in tl_follows.items():
            assert k in target
            assert target[k] == v

    def test_follows_distance_2(self):
        d = {("a", "b", "c"): 2, ("a", "f"): 1}
        s = {("a", "c"): 2}
        tl = TraceLog(d)
        s_tl = tl.follows(distance=2)

        for k, v in s.items():
            assert k in s_tl
            assert s_tl[k] == v

        for k, v in s_tl.items():
            assert k in s
            assert s[k] == v

    
    def test_successors_L1(self):
        tl = TraceLog.from_txt(os.path.join(DATA, "L1.txt"))
        tl = tl.augment()

        target = {
            ("[>", "a1"): 20,
            ("a1", "a2"): 10,
            ("a1", "a3"): 3,
            ("a1", "a4"): 7,
            ("a2", "a4"): 13,
            ("a2", "a5"): 7,
            ("a3", "a4"): 8,
            ("a3", "a5"): 6,
            ("a4", "a2"): 7,
            ("a4", "a3"): 6,
            ("a4", "a5"): 21,
            ("a5", "a6"): 14,
            ("a5", "a7"): 9,
            ("a5", "a8"): 11,
            ("a6", "a2"): 3,
            ("a6", "a3"): 5,
            ("a6", "a4"): 6,
            ("a7", "[]"): 9,
            ("a8", "[]"): 11,
        }

        s_tl = tl.follows()

        for k, v in target.items():
            assert k in s_tl
            assert s_tl[k] == v

        for k, v in s_tl.items():
            assert k in target
            assert target[k] == v
            
    def test_follows_loop_len_1(self):
        d = {("a", "b", "c"): 2, ("a", "a", "f"): 1}
        s = {("a", "b"): 2, ("b", "c"): 2, ("a", "f"): 1, ("a", "a"): 1}
        tl = TraceLog(d)
        s_tl = tl.follows(distance=1)

        for k, v in s.items():
            assert k in s_tl
            assert s_tl[k] == v

        for k, v in s_tl.items():
            assert k in s
            assert s[k] == v

    def test_follows_exception_no_integer_distance(self):
        d = {("a", "b", "c"): 2, ("a", "a", "f"): 1}
        tl = TraceLog(d)
        with pytest.raises(ValueError):
            tl.follows(distance=1.2)

    def test_follows_exception_no_positive_integer(self):
        d = {("a", "b", "c"): 2, ("a", "a", "f"): 1}
        tl = TraceLog(d)
        with pytest.raises(ValueError):
            tl.follows(distance=-3.3)

    def test_equivalence_L1(self):
        tl = TraceLog.from_txt(os.path.join(DATA, "L1.txt"))
        target = {("[]", "[>"), ("a1", "[]"), ("a1", "[>"), ("a4", "a5")}
        tl_aug = tl.augment()
        R_eq = tl_aug.equivalence()

        for el in target:
            assert el in R_eq or el[::-1] in R_eq

        for el in R_eq:
            assert el in target or el[::-1] in target

    def test_equivalence_L2(self):
        tl = TraceLog.from_txt(os.path.join(DATA, "L2.txt"))
        target = {
            ("[]", "[>"),
            ("a", "[]"),
            ("a", "[>"),
            ("d", "[]"),
            ("d", "[>"),
            ("a", "d"),
            ("b", "c"),
            ("e", "f"),
        }
        tl_aug = tl.augment()
        R_eq = tl_aug.equivalence()

        for el in target:
            assert el in R_eq or el[::-1] in R_eq

        for el in R_eq:
            assert el in target or el[::-1] in target

    def test_equivalence_L4(self):
        tl = TraceLog.from_txt(os.path.join(DATA, "L4.txt"))
        target = {("[]", "[>")}
        tl_aug = tl.augment()
        R_eq = tl_aug.equivalence()

        for el in target:
            assert el in R_eq or el[::-1] in R_eq

        for el in R_eq:
            assert el in target or el[::-1] in target

    def test_activity_2_freq(self):
        t = (
            "a1",
            "a2",
            "a4",
            "a5",
            "a6",
            "a3",
            "a4",
            "a5",
            "a6",
            "a4",
            "a3",
            "a5",
            "a6",
            "a2",
            "a4",
            "a5",
            "a7",
        )
        target = {"a1": 1, "a2": 2, "a3": 2, "a4": 4, "a5": 4, "a6": 3, "a7": 1}
        a2f = TraceLog.activity_2_freq(t)
        assert a2f == target

    def test_freq_2_activities(self):
        t = (
            "a1",
            "a2",
            "a4",
            "a5",
            "a6",
            "a3",
            "a4",
            "a5",
            "a6",
            "a4",
            "a3",
            "a5",
            "a6",
            "a2",
            "a4",
            "a5",
            "a7",
        )
        target = {1: {"a1", "a7"}, 2: {"a2", "a3"}, 3: {"a6"}, 4: {"a4", "a5"}}
        f2a = TraceLog.freq_2_activities(t)
        assert f2a == target

    def test_never_together(self):
        d = {
            ("a", "b", "c"): 2,
            ("b", "c", "a"): 1,
            ("a", "c"): 1,
            ("f", "a"): 1,
            ("a",): 2,
        }
        target = {("b", "f"), ("c", "f")}

        tl = TraceLog(d)
        nt = tl.never_together()

        for pair in nt:
            assert pair in target

        for pair in target:
            assert pair in nt

    def test_never_together_L1(self):
        tl = TraceLog.from_txt(os.path.join(DATA, "L1.txt"))
        tl_aug = tl.augment()
        target = {("a7", "a8")}

        nt = tl_aug.never_together()

        for pair in nt:
            assert pair in target

        for pair in target:
            assert pair in nt

    def test_always_after_L1(self):
        tl = TraceLog.from_txt(os.path.join(DATA, "L1.txt"))
        target = {
            ("[>", "a1"),
            ("[>", "a4"),
            ("[>", "a5"),
            ("[>", "[]"),
            ("a1", "a4"),
            ("a1", "a5"),
            ("a1", "[]"),
            ("a2", "a5"),
            ("a2", "[]"),
            ("a3", "[]"),
            ("a3", "a5"),
            ("a4", "a5"),
            ("a4", "[]"),
            ("a5", "[]"),
            ("a6", "a4"),
            ("a6", "a5"),
            ("a6", "[]"),
            ("a7", "[]"),
            ("a8", "[]")
        }
        tl_aug = tl.augment()
        R_aa = tl_aug.always_after()

        for pair in R_aa:
            assert pair in target

        for pair in target:
            assert pair in R_aa

    def test_always_after_xes(self):
        tl = TraceLog.from_xes(os.path.join(DATA, "L2.xes.gz"))
        target = {
            ("[>", "[]"),
            ("[>", "a"),
            ("[>", "b"),
            ("[>", "c"),
            ("[>", "d"),
            ("a", "b"),
            ("a", "c"),
            ("a", "d"),
            ("b", "d"),
            ("c", "d"),
            ("e", "b"),
            ("e", "c"),
            ("e", "d"),
            ("e", "f"),
            ("f", "b"),
            ("f", "c"),
            ("f", "d"),
            ("a", "[]"),
            ("b", "[]"),
            ("c", "[]"),
            ("d", "[]"),
            ("e", "[]"),
            ("f", "[]"),
        }
        tl_aug = tl.augment()
        R_aa = tl_aug.always_after()

        for pair in R_aa:
            assert pair in target

        for pair in target:
            assert pair in R_aa

    def test_always_before_L1(self):
        tl = TraceLog.from_txt(os.path.join(DATA, "L1.txt"))
        target = {
            ("a1", "[>"),
            ("a2", "[>"),
            ("a3", "[>"),
            ("a4", "[>"),
            ("a5", "[>"),
            ("a6", "[>"),
            ("a7", "[>"),
            ("a8", "[>"),
            ("[]", "[>"),
            ("a2", "a1"),
            ("a3", "a1"),
            ("a4", "a1"),
            ("a5", "a1"),
            ("a6", "a1"),
            ("a7", "a1"),
            ("a8", "a1"),
            ("[]", "a1"),
            ("a5", "a4"),
            ("a6", "a4"),
            ("a7", "a4"),
            ("a8", "a4"),
            ("[]", "a4"),
            ("a6", "a5"),
            ("a7", "a5"),
            ("a8", "a5"),
            ("[]", "a5"),
        }
        tl_aug = tl.augment()
        R_ab = tl_aug.always_before()

        for pair in R_ab:
            assert pair in target

        for pair in target:
            assert pair in R_ab

    def test_from_txt(self):
        tl = TraceLog.from_txt(os.path.join(DATA, "L2.txt"))
        target = {
            ("a", "b", "c", "d"): 3,
            ("a", "c", "b", "d"): 4,
            ("a", "b", "c", "e", "f", "b", "c", "d"): 2,
            ("a", "b", "c", "e", "f", "c", "b", "d"): 1,
            ("a", "c", "b", "e", "f", "b", "c", "d"): 2,
            ("a", "c", "b", "e", "f", "b", "c", "e", "f", "c", "b", "d"): 1,
        }

        for k, v in target.items():
            assert k in tl
            assert tl[k] == v

        for k, v in tl.items():
            assert k in target
            assert target[k] == v

    def test_from_xes(self):
        tl = TraceLog.from_xes(os.path.join(DATA, "L2.xes"))
        target = {
            ("a", "b", "c", "d"): 3,
            ("a", "c", "b", "d"): 4,
            ("a", "b", "c", "e", "f", "b", "c", "d"): 2,
            ("a", "b", "c", "e", "f", "c", "b", "d"): 1,
            ("a", "c", "b", "e", "f", "b", "c", "d"): 2,
            ("a", "c", "b", "e", "f", "b", "c", "e", "f", "c", "b", "d"): 1,
        }

        for k, v in target.items():
            assert k in tl
            assert tl[k] == v

        for k, v in tl.items():
            assert k in target
            assert target[k] == v

    def test_from_xes_gunzip(self):
        tl = TraceLog.from_xes(os.path.join(DATA, "L2.xes.gz"))
        target = {
            ("a", "b", "c", "d"): 3,
            ("a", "c", "b", "d"): 4,
            ("a", "b", "c", "e", "f", "b", "c", "d"): 2,
            ("a", "b", "c", "e", "f", "c", "b", "d"): 1,
            ("a", "c", "b", "e", "f", "b", "c", "d"): 2,
            ("a", "c", "b", "e", "f", "b", "c", "e", "f", "c", "b", "d"): 1,
        }

        for k, v in target.items():
            assert k in tl
            assert tl[k] == v

        for k, v in tl.items():
            assert k in target
            assert target[k] == v

    def test_from_txt_exception_duplicate_trace(self):
        with pytest.raises(IllegalLogAction):
            tl = TraceLog.from_txt(os.path.join(DATA, "L2_duplicate_trace.txt"))

    def test_from_txt_exception_invalid_frequency(self):
        with pytest.raises(IllegalLogAction):
            tl = TraceLog.from_txt(os.path.join(DATA, "L2_invalid_frequency.txt"))

    def test_augment(self):
        tl = TraceLog.from_txt(os.path.join(DATA, "L2.txt"))
        target = {
            ("[>", "a", "b", "c", "d", "[]"): 3,
            ("[>", "a", "c", "b", "d", "[]"): 4,
            ("[>", "a", "b", "c", "e", "f", "b", "c", "d", "[]"): 2,
            ("[>", "a", "b", "c", "e", "f", "c", "b", "d", "[]"): 1,
            ("[>", "a", "c", "b", "e", "f", "b", "c", "d", "[]"): 2,
            ("[>", "a", "c", "b", "e", "f", "b", "c", "e", "f", "c", "b", "d", "[]"): 1,
        }
        tl_aug = tl.augment()

        for k, v in target.items():
            assert k in tl_aug
            assert tl_aug[k] == v

        for k, v in tl_aug.items():
            assert k in target
            assert target[k] == v

    def test_sum_counter_L1(self):
        tl = TraceLog.from_txt(os.path.join(DATA, "L1.txt"))
        tl_aug = tl.augment()
        target = {
            "[]": 20,
            "[>": 20,
            "a1": 20,
            "a2": 20,
            "a3": 14,
            "a4": 34,
            "a5": 34,
            "a6": 14,
            "a7": 9,
            "a8": 11,
        }
        
        count = tl_aug.sum_counter()

        for k, v in count.items():
            assert k in target
            assert target[k] == v

        for a, f in target.items():
            assert a in count
            assert count[a] == f
    
    def test_min_counter_L1(self):
        tl = TraceLog.from_txt(os.path.join(DATA, "L1.txt"))
        tl_aug = tl.augment()
        target = {
            "[]": 1,
            "[>": 1,
            "a1": 1,
            "a2": 0,
            "a3": 0,
            "a4": 1,
            "a5": 1,
            "a6": 0,
            "a7": 0,
            "a8": 0,
        }
        count = tl_aug.min_counter()

        for k, v in count.items():
            assert k in target
            assert target[k] == v

        for a, f in target.items():
            assert a in count
            assert count[a] == f

    def test_max_counter_L1(self):
        tl = TraceLog.from_txt(os.path.join(DATA, "L1.txt"))
        tl_aug = tl.augment()
        target = {
            "[]": 1,
            "[>": 1,
            "a1": 1,
            "a2": 3,
            "a3": 2,
            "a4": 4,
            "a5": 4,
            "a6": 3,
            "a7": 1,
            "a8": 1,
        }
        count = tl_aug.max_counter()

        for k, v in count.items():
            assert k in target
            assert target[k] == v

        for a, f in target.items():
            assert a in count
            assert count[a] == f

    def test_sum_counter_L4(self):
        tl = TraceLog.from_txt(os.path.join(DATA, "L4.txt"))
        tl_aug = tl.augment()
        target = {
            "[]": 255,
            "[>": 255,
            "a": 128,
            "b": 127,
            "c": 363,
            "d": 174,
            "e": 81,
        }
        
        count = tl_aug.sum_counter()
        print(count)

        for k, v in count.items():
            assert k in target
            assert target[k] == v

        for a, f in target.items():
            assert a in count
            assert count[a] == f
    
    def test_min_counter_L4(self):
        tl = TraceLog.from_txt(os.path.join(DATA, "L4.txt"))
        tl_aug = tl.augment()
        target = {
            "[]": 1,
            "[>": 1,
            "a": 0,
            "b": 0,
            "c": 1,
            "d": 0,
            "e": 0,
        }
        count = tl_aug.min_counter()

        for k, v in count.items():
            assert k in target
            assert target[k] == v

        for a, f in target.items():
            assert a in count
            assert count[a] == f

    def test_max_counter_L4(self):
        tl = TraceLog.from_txt(os.path.join(DATA, "L4.txt"))
        tl_aug = tl.augment()
        target = {
            "[]": 1,
            "[>": 1,
            "a": 1,
            "b": 1,
            "c": 2,
            "d": 1,
            "e": 1,
        }
        count = tl_aug.max_counter()

        for k, v in count.items():
            assert k in target
            assert target[k] == v

        for a, f in target.items():
            assert a in count
            assert count[a] == f

    def test_filter_traces(self):
        tl = TraceLog.from_xes(os.path.join(DATA, "L2.xes"))
        
        reqA = {'a', 'b', 'c'}
        forbA = {'e', 'f'}

        # target = {
        #     ('a', 'b', 'c', 'e', 'f', 'b', 'c', 'd') : 2,
        #     ('a', 'b', 'c', 'e', 'f', 'c', 'b', 'd') : 1,
        #     ('a', 'c', 'b', 'e', 'f', 'b', 'c', 'd') : 2,
        #     ('a', 'c', 'b', 'e', 'f', 'b', 'c', 'e', 'f', 'c', 'b', 'd') : 1
        # }

        target = {
            ('a', 'b', 'c', 'd') : 3,
            ('a', 'c', 'b', 'd') : 4
        }

        fa = tl.filter_traces(reqA, forbA)

        for k, v in fa.items():
            assert k, v in target
        
        for k, v in tl.items():
            assert k, v in fa


from itertools import combinations
from sortedcontainers import SortedSet

def follows(trace, distance=1):
    if not isinstance(trace, tuple):
        raise ValueError("Trace has to be a tuple of activities.")
    if not float(distance).is_integer():
        raise ValueError("Distance has to be an integer.")
    if not distance >= 1:
        raise ValueError("Distance has to be greater or equal to 1.")

    pairs = dict()

    for i in range(len(trace) - distance):
        ai = trace[i]
        aj = trace[i + distance]

        if (ai, aj) not in pairs:
            pairs[(ai, aj)] = 0

        pairs[(ai, aj)] += 1

    return pairs

def successors(trace):
    if not isinstance(trace, tuple):
        raise ValueError("Trace has to be a tuple of activities.")

    s = dict()

    for pair in combinations(trace, r=2):

        if pair[0] not in s:
            s[pair[0]] = SortedSet()

        s[pair[0]].add(pair[1])

    return s

def predecessors(trace):
    if not isinstance(trace, tuple):
        raise ValueError("Trace has to be a tuple of activities.")

    return successors(tuple(reversed(trace)))


import itertools, re
from collections.abc import MutableMapping
import xml.etree.ElementTree as etree
from lxml import etree as etree2
from copy import deepcopy
import gzip
from io import BytesIO

from pm4py.objects.log.importer.xes import factory as xes_import_factory
from sortedcontainers import SortedSet

from .exceptions import IllegalLogAction
from .utils import follows, predecessors, successors


class TraceLog(MutableMapping):
    """Representation of a trace log. Works like a base python dict,
    where the keys are tuples denoting individual traces
    (e.g. '("a", "b", "c")' denoted trace 'abc') and the values
    denote the frequencies of the traces.
    """

    def __init__(self, *args, **kwargs):
        self.__traces = dict()
        self.__traces.update(*args, **kwargs)
        self.__labels = SortedSet()

        for trace in self.__traces:
            for activity in trace:
                self.__labels.add(activity)

    def __setitem__(self, key, value):
        if not float(value).is_integer() or value < 0:
            raise IllegalLogAction(
                "Cannot set value at key {} equal to {}.".format(key, value)
            )
        self.__traces[key] = value
        # If there is a new activity add it to the set of labels
        for activity in key:
            self.__labels.add(activity)

    def __getitem__(self, key):
        return self.__traces[key]

    def __delitem__(self, key):
        del self.__traces[key]

    def __iter__(self):
        return iter(self.__traces)

    def __len__(self):
        return len(self.__traces)

    def __str__(self):
        """returns simple dict representation of the mapping"""
        return str(self.__traces)

    def __repr__(self):
        """echoes class, id, & reproducible representation in the REPL"""
        return "{}, D({})".format(super(TraceLog, self).__repr__(), self.__traces)

    @property
    def labels(self):
        """Returns all the unique labels of activities in the trace log."""
        return self.__labels

    def augment(self, start="[>", end="[]"):
        """Returns a similar TraceLog object where each trace contains an aditional
        start and end activity
        """
        tl = TraceLog()
        for key, value in self.__traces.items():
            trace = (start,) + key + (end,)
            tl[trace] = value
        return tl

    def follows(self, distance=1):
        """Returns a mapping (aka. dict) from pairs of activities to frequency.
        A pair (a, b) is part of the mapping if activity b follows activity a,
        at a certain distance, in any of the traces.

        Parameters
        ----------
        distance: int
            Distance two activities have to be appart to be counted in the mapping.
        """
        pairs = dict()

        for trace in self.__traces:
            # Get the follows mapping only for the current trace
            f = follows(trace, distance=distance)

            # Add all the items to the overall dictionary
            for p, p_freq in f.items():
                # If it's not there yet, add the default value
                if p not in pairs:
                    pairs[p] = 0
                pairs[p] += p_freq * self.__traces[trace]

        return pairs

    def save_to_file(self, filepath, format="txt"):
        """Save a TraceLog object as a `.txt` file.
        """
        if len(self.__traces) == 0:
            return False

        output = ""

        if format == "txt":
            for i, kv in enumerate(self.__traces.items()):
                key = kv[0]
                value = kv[1]
                output += "{}x Case{} {}\n".format(value, i, " ".join(key))

        with open(filepath, "w") as f:
            f.write(output)

        return True

    def never_together(self):
        """Returns a set of tuples, representing the pairs of the activities
        which are never together in any of the traces.

        Returns
        -------
        `set` of `tuples`
            the pairs of the activities which are never together in any of the traces
        """

        pairs = set(itertools.combinations(self.labels, r=2))

        for trace in self.__traces:
            trace = SortedSet(trace)
            trace_pairs = list(itertools.combinations(trace, r=2))
            pairs_wc = deepcopy(pairs)

            for pair in pairs:
                if pair in trace_pairs:
                    pairs_wc.discard(pair)

            pairs = pairs_wc

        return pairs

    def equivalence(self):
        """Returns a set of tuples, representing the pairs of the activities
        which are always together in all of the traces the same number of times.

        Returns
        -------
        `set``of `tuples`
            the pairs of the activities which are always together in all of the
            traces the same number of times
        """
        R_eq_trace = dict()

        for trace in self.__traces:
            # Extract pairs in current trace
            w = dict()
            f2a = TraceLog.freq_2_activities(trace)
            for s in f2a.values():
                for pair in itertools.product(s, repeat=2):
                    if pair[0] == pair[1]:
                        continue
                    if pair[0] not in w:
                        w[pair[0]] = list()
                    w[pair[0]].append(pair[1])

            # Check if the relationships found till now still hold
            for pair, value in R_eq_trace.items():
                if value:
                    if pair[0] not in w and pair[1] not in w:
                        continue
                    if pair[0] in w and pair[1] not in w[pair[0]]:
                        R_eq_trace[pair] = False
                        R_eq_trace[pair[::-1]] = False

            # Check if there are new values to be added
            for v0, values in w.items():
                for v1 in values:
                    if (v0, v1) not in R_eq_trace:
                        R_eq_trace[(v0, v1)] = True

        # Transform the dict to set
        R_eq = set()
        for pair, value in R_eq_trace.items():
            if value:
                R_eq.add(pair)

        return R_eq

    def always_after(self):
        """Returns a set of tuples, representing the pairs of the activities
        which after any occurrence of the first activity the second activity always occurs.

        Returns
        -------
        `set``of `tuples`
            pairs of the activities which after any occurrence of the first activity the
            second activity always occurs.
        """
        pairs = set(itertools.permutations(self.labels, r=2))

        # Remove impossible pairs
        first = "[>"
        last = "[]"
        pairs = set([(a, b) for a, b in pairs if (a, b) != (a, first) and (a, b) != (last, b)])

        for trace in self.__traces:
            s = successors(trace)

            pairs_wc = set([(a, b) for a, b in pairs if a in s.keys() and b not in s[a]])
            pairs = pairs - pairs_wc

        return pairs

    def always_before(self):
        """Returns a set of tuples, representing the pairs of the activities
        which before any occurrence of the first activity the second activity always occurs.

        Returns
        -------
        `set``of `tuples`
            pairs of the activities which before any occurrence of the first activity the
            second activity always occurs.
        """
        pairs = set(itertools.permutations(self.labels, r=2))

        # remove impossible pairs
        first = "[>"
        last = "[]"
        pairs = set([(a, b) for a, b in pairs if (a, b) != (first, b) and (a, b) != (a, last)])

        for trace in self.__traces:
            p = predecessors(trace)

            pairs_wc = set([(a, b) for a, b in pairs if a in p.keys() and b not in p[a]])
            pairs = pairs - pairs_wc

        return pairs

    @staticmethod
    def activity_2_freq(trace):
        """For a given trace, return a mapping from activity to frequency in trace.

        Parameters
        ----------
        trace: `tuple` of `str`
            a trace as a tuple of activities

        Returns
        -------
        `dict`
            mapping from activity to frequency in trace
        """
        d = {}
        for a in trace:
            if a not in d:
                d[a] = 0
            d[a] += 1

        return d

    @staticmethod
    def freq_2_activities(trace):
        """For a given trace, return a mapping from frequency to set of activities,
        with that frequency in the trace.

        Parameters
        ----------
        trace: `tuple` of `str`
            a trace as a tuple of activities

        Returns
        -------
        `dict`
            mapping from frequency to `set` activities in trace
        """
        a2f = TraceLog.activity_2_freq(trace)
        f2a = {}
        for key, value in a2f.items():
            if value not in f2a:
                f2a[value] = set()
            f2a[value].add(key)
        return f2a

    def sum_counter(self):
        """Returns a dict, representing a Mapping from activity to the amount of times the activity
        appears in the TraceLog.

         Parameters
        ----------
        activity: 'str'
            the name of the activity

        Returns
        -------
        'dict'
            Mapping from activity to the amount of times the activity appears in the TraceLog.
        """
        sum_c = dict()

        for trace in self.__traces:
            cur = self.activity_2_freq(trace)
            for k, v in cur.items():
                if k not in sum_c:
                    sum_c[k] = 0
                sum_c[k] += v * self.__traces[trace]

        return sum_c

    def min_counter(self):
        """Returns a dict, representing a Mapping from activity to the min amount of times the
        activity appears in any trace of the TraceLog.

         Parameters
        ----------
        activity: 'str'
            the name of the activity

        Returns
        -------
        'dict'
            Mapping from activity to the min amount of times the activity appears in any trace of the TraceLog.
        """
        base = dict()
        for a in self.labels:
            base[a] = 0
        min_c = dict()
        for trace in self.__traces:
            cur = deepcopy(base)
            cur.update(self.activity_2_freq(trace))
            for k, v in cur.items():
                if k not in min_c:
                    min_c[k] = v
                elif v < min_c[k]:
                    min_c[k] = v
        return min_c

    def max_counter(self):
        """Returns a dict, representing a Mapping from activity to the max amount of times the
        activity appears in any trace of the TraceLog.

         Parameters
        ----------
        activity: 'str'
            the name of the activity

        Returns
        -------
        'dict'
            Mapping from activity to the max amount of times the activity appears in any trace of the TraceLog.
        """
        max_c = dict()
        for trace in self.__traces:
            cur = self.activity_2_freq(trace)
            for k, v in cur.items():
                if k not in max_c:
                    max_c[k] = v
                elif v > max_c[k]:
                    max_c[k] = v
        return max_c

    def filter_traces(self, reqA, forbA):
        """Filters the tracelog based on required and forbidden activities.

        Parameters
        ----------
        reqA: `set()`
            If one or more of the selected activities
            does not occur in a trace, the entire trace will be filtered out.
        forbA: `set()`
            If one or more of the selected activities
            occurs in a trace, the entire trace will be filtered out.

        Returns
        -------
        `TraceLog`
            Mapping from activity to coresponding event list.
        """

        if reqA:
            for trace in list(self.__traces):
                for a in reqA:
                    if a not in trace:
                        self.__traces.__delitem__(trace)
                        break

        if forbA:
            for trace in list(self.__traces):
                for a in forbA:
                    if a in trace:
                        self.__traces.__delitem__(trace)
                        break

        return self.__traces

    @staticmethod
    def from_txt(filepath, delimiter=None, frequency_idx=0, first_activity_idx=2):
        """Parses a `.txt` file containing a trace log and returns a TraceLog object of it.

        Parameters
        ----------
        filepath: path-like
            The path to the `.txt` file.
        delimiter: `str`
            Character delimiting the different values. Default None, thus splitting by all the whitespace.
        frequency_idx: `int`
            Default 0.
        first_activity_idx: `int`
            Default 2.

        Returns
        -------
        `TraceLog`
            Mapping from activity to coresponding event list.
        """
        tl = TraceLog()

        with open(filepath, "r") as f:
            for row in f:

                row = row.strip()
                if len(row) == 0:
                    continue

                parts = row.split(delimiter)
                a = tuple(parts[first_activity_idx:])
                try:
                    frequency = int((parts[frequency_idx]).replace("x", ""))
                except Exception:
                    raise IllegalLogAction("No frequency for trace: {}.".format(a))

                if a in tl:
                    raise IllegalLogAction(
                        "Attempting to add trace {} twice.".format(a)
                    )

                tl[a] = frequency

        return tl

    @staticmethod
    def from_xes(filepath):
        file_context = filepath

        if filepath.endswith(".gz"):
            with gzip.open(filepath, "r") as f:
                file_context = BytesIO(f.read())

        context = etree2.iterparse(file_context, events=["start", "end"])

        in_event = False
        tracelog = dict()

        for tree_event, elem in context:
            if elem.tag.endswith("trace"):
                if tree_event == "start":
                    trace = ()
                else:
                    if trace not in tracelog:
                        tracelog[trace] = 0
                    tracelog[trace] += 1

            if elem.tag.endswith("event"):
                in_event = tree_event == "start"

            if (
                in_event
                and "key" in elem.attrib
                and elem.attrib["key"] == "concept:name"
            ):
                if tree_event == "start":
                    trace += (elem.attrib["value"],)
            
            elem.clear()

        return TraceLog(tracelog)

